# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/common/action/get_enum_attribute.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/core/node/common/action/get_enum_attribute.proto",
    package="syft.core.node.common.action",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n6proto/core/node/common/action/get_enum_attribute.proto\x12\x1csyft.core.node.common.action\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\xb9\x01\n\x16GetEnumAttributeAction\x12\x0c\n\x04path\x18\x01 \x01(\t\x12-\n\x0eid_at_location\x18\x02 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address\x12%\n\x06msg_id\x18\x04 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bvalue_index\x18\x05 \x01(\x05\x62\x06proto3',
    dependencies=[
        proto_dot_core_dot_common_dot_common__object__pb2.DESCRIPTOR,
        proto_dot_core_dot_io_dot_address__pb2.DESCRIPTOR,
    ],
)


_GETENUMATTRIBUTEACTION = _descriptor.Descriptor(
    name="GetEnumAttributeAction",
    full_name="syft.core.node.common.action.GetEnumAttributeAction",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="path",
            full_name="syft.core.node.common.action.GetEnumAttributeAction.path",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="id_at_location",
            full_name="syft.core.node.common.action.GetEnumAttributeAction.id_at_location",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.core.node.common.action.GetEnumAttributeAction.address",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.core.node.common.action.GetEnumAttributeAction.msg_id",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="value_index",
            full_name="syft.core.node.common.action.GetEnumAttributeAction.value_index",
            index=4,
            number=5,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=157,
    serialized_end=342,
)

_GETENUMATTRIBUTEACTION.fields_by_name[
    "id_at_location"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_GETENUMATTRIBUTEACTION.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_GETENUMATTRIBUTEACTION.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
DESCRIPTOR.message_types_by_name["GetEnumAttributeAction"] = _GETENUMATTRIBUTEACTION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GetEnumAttributeAction = _reflection.GeneratedProtocolMessageType(
    "GetEnumAttributeAction",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETENUMATTRIBUTEACTION,
        "__module__": "proto.core.node.common.action.get_enum_attribute_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.common.action.GetEnumAttributeAction)
    },
)
_sym_db.RegisterMessage(GetEnumAttributeAction)


# @@protoc_insertion_point(module_scope)



# stdlib
import secrets
from typing import List
from typing import Optional
from typing import Type
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey
from typing_extensions import final

# syft relative
from ... import serialize
from ...core.common.message import ImmediateSyftMessageWithReply
from ...core.common.message import ImmediateSyftMessageWithoutReply
from ...core.common.message import SyftMessage
from ...core.common.serde.deserialize import _deserialize
from ...core.common.serde.serializable import bind_protobuf
from ...core.common.uid import UID
from ...core.io.address import Address
from ...core.node.abstract.node import AbstractNode
from ...core.node.common.metadata import Metadata
from ...core.node.common.service.auth import service_auth
from ...core.node.common.service.node_service import ImmediateNodeServiceWithReply
from ...core.node.common.service.node_service import ImmediateNodeServiceWithoutReply
from ...proto.grid.service.signaling_service_pb2 import (
    AnswerPullRequestMessage as AnswerPullRequestMessage_PB,
)
from ...proto.grid.service.signaling_service_pb2 import (
    CloseConnectionMessage as CloseConnectionMessage_PB,
)
from ...proto.grid.service.signaling_service_pb2 import (
    InvalidLoopBackRequest as InvalidLoopBackRequest_PB,
)
from ...proto.grid.service.signaling_service_pb2 import (
    OfferPullRequestMessage as OfferPullRequestMessage_PB,
)
from ...proto.grid.service.signaling_service_pb2 import (
    PeerSuccessfullyRegistered as PeerSuccessfullyRegistered_PB,
)
from ...proto.grid.service.signaling_service_pb2 import (
    RegisterNewPeerMessage as RegisterNewPeerMessage_PB,
)
from ...proto.grid.service.signaling_service_pb2 import (
    SignalingAnswerMessage as SignalingAnswerMessage_PB,
)
from ...proto.grid.service.signaling_service_pb2 import (
    SignalingOfferMessage as SignalingOfferMessage_PB,
)
from ...proto.grid.service.signaling_service_pb2 import (
    SignalingRequestsNotFound as SignalingRequestsNotFound_PB,
)


@bind_protobuf
@final
class OfferPullRequestMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        target_peer: str,
        host_peer: str,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.target_peer = target_peer
        self.host_peer = host_peer

    def _object2proto(self) -> OfferPullRequestMessage_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: OfferPullRequestMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return OfferPullRequestMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            target_peer=self.target_peer,
            host_peer=self.host_peer,
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(proto: OfferPullRequestMessage_PB) -> "OfferPullRequestMessage":
        """Creates a OfferPullRequestMessage from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of SignalingOfferMessage
        :rtype: OfferPullRequestMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return OfferPullRequestMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            target_peer=proto.target_peer,
            host_peer=proto.host_peer,
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return OfferPullRequestMessage_PB


@bind_protobuf
@final
class AnswerPullRequestMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        target_peer: str,
        host_peer: str,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.target_peer = target_peer
        self.host_peer = host_peer

    def _object2proto(self) -> AnswerPullRequestMessage_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: AnswerPullRequestMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return AnswerPullRequestMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            target_peer=self.target_peer,
            host_peer=self.host_peer,
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(proto: AnswerPullRequestMessage_PB) -> "AnswerPullRequestMessage":
        """Creates a AnswerPullRequestMessage from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of AnswerPullRequestMessage
        :rtype: AnswerPullRequestMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return AnswerPullRequestMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            target_peer=proto.target_peer,
            host_peer=proto.host_peer,
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return AnswerPullRequestMessage_PB


@bind_protobuf
@final
class RegisterNewPeerMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)

    def _object2proto(self) -> RegisterNewPeerMessage_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: AnswerPullRequestMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return RegisterNewPeerMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(proto: RegisterNewPeerMessage_PB) -> "RegisterNewPeerMessage":
        """Creates a AnswerPullRequestMessage from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of AnswerPullRequestMessage
        :rtype: AnswerPullRequestMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return RegisterNewPeerMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return RegisterNewPeerMessage_PB


@bind_protobuf
@final
class PeerSuccessfullyRegistered(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        peer_id: str,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.peer_id = peer_id

    def _object2proto(self) -> PeerSuccessfullyRegistered_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return PeerSuccessfullyRegistered_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            peer_id=self.peer_id,
        )

    @staticmethod
    def _proto2object(
        proto: PeerSuccessfullyRegistered_PB,
    ) -> "PeerSuccessfullyRegistered":
        """Creates a SignalingOfferMessage from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return PeerSuccessfullyRegistered(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            peer_id=proto.peer_id,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return PeerSuccessfullyRegistered_PB


@bind_protobuf
@final
class SignalingOfferMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        payload: str,
        host_metadata: Metadata,
        target_peer: str,
        host_peer: str,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.payload = payload
        self.host_metadata = host_metadata
        self.target_peer = target_peer
        self.host_peer = host_peer

    def _object2proto(self) -> SignalingOfferMessage_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return SignalingOfferMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            payload=self.payload,
            host_metadata=serialize(self.host_metadata),
            target_peer=self.target_peer,
            host_peer=self.host_peer,
        )

    @staticmethod
    def _proto2object(proto: SignalingOfferMessage_PB) -> "SignalingOfferMessage":
        """Creates a SignalingOfferMessage from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return SignalingOfferMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            payload=proto.payload,
            host_metadata=_deserialize(blob=proto.host_metadata),
            target_peer=proto.target_peer,
            host_peer=proto.host_peer,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return SignalingOfferMessage_PB


@bind_protobuf
@final
class SignalingAnswerMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        payload: str,
        host_metadata: Metadata,
        target_peer: str,
        host_peer: str,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.payload = payload
        self.host_metadata = host_metadata
        self.target_peer = target_peer
        self.host_peer = host_peer

    def _object2proto(self) -> SignalingAnswerMessage_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: SignalingAnswerMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """

        return SignalingAnswerMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            payload=self.payload,
            host_metadata=serialize(self.host_metadata),
            target_peer=self.target_peer,
            host_peer=self.host_peer,
        )

    @staticmethod
    def _proto2object(proto: SignalingAnswerMessage_PB) -> "SignalingAnswerMessage":
        """Creates a SignalingAnswerMessage from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of SignalingAnswerMessage
        :rtype: SignalingAnswerMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return SignalingAnswerMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            payload=proto.payload,
            host_metadata=_deserialize(blob=proto.host_metadata),
            target_peer=proto.target_peer,
            host_peer=proto.host_peer,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return SignalingAnswerMessage_PB


@bind_protobuf
@final
class SignalingRequestsNotFound(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)

    def _object2proto(self) -> SignalingRequestsNotFound_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: SignalingRequestsNotFound_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return SignalingRequestsNotFound_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
        )

    @staticmethod
    def _proto2object(
        proto: SignalingRequestsNotFound_PB,
    ) -> "SignalingRequestsNotFound":
        """Creates a SignalingRequestsNotFound from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of SignalingRequestsNotFound
        :rtype: SignalingRequestsNotFound

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return SignalingRequestsNotFound(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return SignalingRequestsNotFound_PB


@bind_protobuf
@final
class InvalidLoopBackRequest(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)

    def _object2proto(self) -> InvalidLoopBackRequest_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: InvalidLoopBackRequest_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return InvalidLoopBackRequest_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
        )

    @staticmethod
    def _proto2object(
        proto: InvalidLoopBackRequest_PB,
    ) -> "InvalidLoopBackRequest":
        """Creates a InvalidLoopBackRequest from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of InvalidLoopBackRequest
        :rtype: InvalidLoopBackRequest

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return InvalidLoopBackRequest(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return InvalidLoopBackRequest_PB


@final
class CloseConnectionMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)

    def _object2proto(self) -> CloseConnectionMessage_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: CloseConnectionMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return CloseConnectionMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
        )

    @staticmethod
    def _proto2object(
        proto: CloseConnectionMessage_PB,
    ) -> "CloseConnectionMessage":
        """Creates a InvalidLoopBackRequest from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of CloseConnectionMessage
        :rtype: CloseConnectionMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return CloseConnectionMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return CloseConnectionMessage_PB


class RegisterDuetPeerService(ImmediateNodeServiceWithReply):
    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: AbstractNode,
        msg: Union[RegisterNewPeerMessage],
        verify_key: VerifyKey,
    ) -> PeerSuccessfullyRegistered:
        peer_id = secrets.token_hex(nbytes=16)
        node.signaling_msgs[peer_id] = {VerifyKey: verify_key, SyftMessage: {}}
        return PeerSuccessfullyRegistered(address=msg.reply_to, peer_id=peer_id)

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateSyftMessageWithReply]]:
        return [RegisterNewPeerMessage]


class PushSignalingService(ImmediateNodeServiceWithoutReply):
    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: AbstractNode,
        msg: Union[SignalingOfferMessage, SignalingAnswerMessage],
        verify_key: VerifyKey,
    ) -> None:
        _peer_signaling = node.signaling_msgs.get(msg.target_peer, None)

        # Do not store loopback signaling requests
        if msg.host_peer != msg.target_peer and _peer_signaling:
            # TODO: remove hacky signaling_msgs when SyftMessages become Storable.
            _peer_signaling[SyftMessage][msg.id] = msg

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateSyftMessageWithoutReply]]:
        return [SignalingOfferMessage, SignalingAnswerMessage]


class PullSignalingService(ImmediateNodeServiceWithReply):

    _pull_push_mapping = {
        OfferPullRequestMessage: SignalingOfferMessage,
        AnswerPullRequestMessage: SignalingAnswerMessage,
    }

    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: AbstractNode,
        msg: Union[OfferPullRequestMessage, AnswerPullRequestMessage],
        verify_key: VerifyKey,
    ) -> Union[
        SignalingOfferMessage,
        SignalingAnswerMessage,
        SignalingRequestsNotFound,
        InvalidLoopBackRequest,
    ]:

        # Do not allow loopback signaling requests
        if msg.host_peer == msg.target_peer:
            return InvalidLoopBackRequest(address=msg.reply_to)

        try:
            sig_requests_for_me = (
                lambda push_msg: push_msg.target_peer == msg.host_peer
                and push_msg.host_peer == msg.target_peer
                and node.signaling_msgs[msg.host_peer][VerifyKey] == verify_key
                and isinstance(
                    push_msg, PullSignalingService._pull_push_mapping[type(msg)]
                )
            )

            # TODO: remove hacky signaling_msgs when SyftMessages become Storable.
            results = list(
                filter(
                    sig_requests_for_me,
                    node.signaling_msgs[msg.host_peer][SyftMessage].values(),
                )
            )

            if results:
                result_msg = results.pop(0)  # FIFO

                # TODO: remove hacky signaling_msgs when SyftMessages become Storable.
                return node.signaling_msgs[msg.host_peer][SyftMessage].pop(
                    result_msg.id
                )  # Retrieve and remove it from storage
            else:
                return SignalingRequestsNotFound(address=msg.reply_to)
        except KeyError:
            return SignalingRequestsNotFound(address=msg.reply_to)

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateSyftMessageWithReply]]:
        return [OfferPullRequestMessage, AnswerPullRequestMessage]



# stdlib
from collections import OrderedDict
from typing import Iterable
from typing import KeysView
from typing import Optional
from typing import ValuesView

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft relative
from . import ObjectStore
from ...logger import critical
from ...logger import traceback_and_raise
from ..common.uid import UID
from .storeable_object import StorableObject


class MemoryStore(ObjectStore):
    """
    Class that implements an in-memory ObjectStorage, backed by a dict.

    Attributes:
        _objects (dict): the dict that backs the storage of the MemoryStorage.
        _search_engine (ObjectSearchEngine): the objects that handles searching by using tags or
        description.
    """

    __slots__ = ["_objects", "_search_engine"]

    def __init__(self) -> None:
        super().__init__()
        self._objects: OrderedDict[UID, StorableObject] = OrderedDict()
        self._search_engine = None
        self.post_init()

    def get_object(self, key: UID) -> Optional[StorableObject]:
        return self._objects.get(key, None)

    def get_objects_of_type(self, obj_type: type) -> Iterable[StorableObject]:
        return [obj for obj in self.values() if isinstance(obj.data, obj_type)]

    def __sizeof__(self) -> int:
        return self._objects.__sizeof__()

    def __str__(self) -> str:
        return str(self._objects)

    def __len__(self) -> int:
        return len(self._objects)

    def keys(self) -> KeysView[UID]:
        return self._objects.keys()

    def values(self) -> ValuesView[StorableObject]:
        return self._objects.values()

    def __contains__(self, key: UID) -> bool:
        return key in self._objects.keys()

    def __getitem__(self, key: UID) -> StorableObject:
        try:
            return self._objects[key]
        except Exception as e:
            critical(f"{type(self)} __getitem__ error {key} {e}")
            traceback_and_raise(e)

    def __setitem__(self, key: UID, value: StorableObject) -> None:
        self._objects[key] = value

    def delete(self, key: UID) -> None:
        try:
            obj = self.get_object(key=key)
            if obj is not None:
                self._objects.__delitem__(key)
            else:
                critical(f"{type(self)} __delitem__ error {key}.")
        except Exception as e:
            critical(f"{type(self)} Exception in __delitem__ error {key}. {e}")

    def clear(self) -> None:
        self._objects.clear()

    def _object2proto(self) -> GeneratedProtocolMessageType:
        pass

    @staticmethod
    def _proto2object(proto: GeneratedProtocolMessageType) -> "MemoryStore":
        pass

    def __repr__(self) -> str:
        return self._objects.__repr__()



# stdlib
from typing import Optional
from urllib.parse import ParseResult
from urllib.parse import urlparse

# syft relative
from ..logger import traceback_and_raise
from .fl_job import FLJob
from .model_centric_fl_worker import ModelCentricFLWorker


class FLClient:
    def __init__(self, url: str, auth_token: str, secure: bool = True) -> None:
        self.url = url
        self.auth_token = auth_token
        self.worker_id: Optional[
            str
        ] = None  # UUID like 4b65bb0f-9530-45b7-a410-482c68c5162e
        url_fragments: ParseResult = urlparse(url)
        if not url_fragments.netloc:
            traceback_and_raise(
                ValueError(
                    f"Cannot create {self} with url: {url}, {url_fragments}. No netloc."
                )
            )
        self.grid_worker = ModelCentricFLWorker(
            address=url_fragments.netloc,
            secure=secure,
        )

    def new_job(self, model_name: str, model_version: str) -> FLJob:
        if self.worker_id is None:
            try:
                auth_response = self.grid_worker.authenticate(
                    self.auth_token, model_name, model_version
                )
                if "data" not in auth_response or "worker_id" in auth_response["data"]:
                    raise traceback_and_raise(
                        ValueError("Missing keys in json response")
                    )
                self.worker_id = auth_response["data"]["worker_id"]
            except Exception as e:
                traceback_and_raise(
                    ValueError(
                        f"Failed to get worker_id from Worker: {self.grid_worker}. {e}"
                    )
                )

        job = FLJob(
            worker_id=self.worker_id,
            grid_worker=self.grid_worker,
            model_name=model_name,
            model_version=model_version,
        )
        return job



# stdlib
from typing import Optional

# syft relative
from . import collections
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ...core.node.abstract.node import AbstractNodeClient
from ..misc.union import UnionGenerator
from .bool import Bool
from .complex import Complex
from .dict import Dict
from .float import Float
from .int import Int
from .iterator import Iterator
from .list import List
from .none import SyNone
from .none import _SyNone
from .primitive_container import Any
from .primitive_interface import PyPrimitive
from .set import Set
from .slice import Slice
from .string import String
from .tuple import Tuple

for syft_type in [
    Bool,
    Complex,
    Dict,
    Float,
    Int,
    SyNone,
    _SyNone,
    Any,
    PyPrimitive,
    Slice,
    String,
    Tuple,
]:
    syft_type.__module__ = __name__


def create_python_ast(client: Optional[AbstractNodeClient] = None) -> Globals:
    ast = Globals(client)

    modules = ["syft", "syft.lib", "syft.lib.python", "syft.lib.python.collections"]
    classes = [
        ("syft.lib.python.Bool", "syft.lib.python.Bool", Bool),
        ("syft.lib.python.Complex", "syft.lib.python.Complex", Complex),
        ("syft.lib.python.Dict", "syft.lib.python.Dict", Dict),
        ("syft.lib.python.Float", "syft.lib.python.Float", Float),
        ("syft.lib.python.Int", "syft.lib.python.Int", Int),
        ("syft.lib.python.List", "syft.lib.python.List", List),
        ("syft.lib.python.Slice", "syft.lib.python.Slice", Slice),
        ("syft.lib.python.String", "syft.lib.python.String", String),
        ("syft.lib.python._SyNone", "syft.lib.python._SyNone", _SyNone),
        ("syft.lib.python.PyPrimitive", "syft.lib.python.PyPrimitive", PyPrimitive),
        ("syft.lib.python.Any", "syft.lib.python.Any", Any),
        ("syft.lib.python.Tuple", "syft.lib.python.Tuple", Tuple),
        ("syft.lib.python.Iterator", "syft.lib.python.Iterator", Iterator),
        ("syft.lib.python.Set", "syft.lib.python.Set", Set),
        (
            "syft.lib.python.collections.OrderedDict",
            "syft.lib.python.collections.OrderedDict",
            collections.OrderedDict,
        ),
    ]

    methods = [
        # Slice methods - quite there
        ("syft.lib.python.Slice.__eq__", "syft.lib.python.Bool"),
        ("syft.lib.python.Slice.__ge__", "syft.lib.python.Bool"),
        ("syft.lib.python.Slice.__gt__", "syft.lib.python.Bool"),
        ("syft.lib.python.Slice.__le__", "syft.lib.python.Bool"),
        ("syft.lib.python.Slice.__lt__", "syft.lib.python.Bool"),
        ("syft.lib.python.Slice.__ne__", "syft.lib.python.Bool"),
        ("syft.lib.python.Slice.__repr__", "syft.lib.python.String"),
        ("syft.lib.python.Slice.__str__", "syft.lib.python.String"),
        ("syft.lib.python.Slice.indices", "syft.lib.python.Tuple"),
        (
            "syft.lib.python.Slice.start",
            UnionGenerator["syft.lib.python.Int", "syft.lib.python._SyNone"],
        ),
        (
            "syft.lib.python.Slice.step",
            UnionGenerator["syft.lib.python.Int", "syft.lib.python._SyNone"],
        ),
        (
            "syft.lib.python.Slice.stop",
            UnionGenerator["syft.lib.python.Int", "syft.lib.python._SyNone"],
        ),
        # List methods - quite there
        ("syft.lib.python.List.__len__", "syft.lib.python.Int"),
        ("syft.lib.python.List.__getitem__", "syft.lib.python.Any"),
        ("syft.lib.python.List.__iter__", "syft.lib.python.Iterator"),
        ("syft.lib.python.List.__add__", "syft.lib.python.List"),
        ("syft.lib.python.List.append", "syft.lib.python._SyNone"),
        ("syft.lib.python.List.__gt__", "syft.lib.python.Bool"),
        ("syft.lib.python.List.__lt__", "syft.lib.python.Bool"),
        ("syft.lib.python.List.__le__", "syft.lib.python.Bool"),
        ("syft.lib.python.List.__ge__", "syft.lib.python.Bool"),
        ("syft.lib.python.List.__iadd__", "syft.lib.python.List"),
        ("syft.lib.python.List.__imul__", "syft.lib.python.List"),
        ("syft.lib.python.List.__iadd__", "syft.lib.python.List"),
        ("syft.lib.python.List.__contains__", "syft.lib.python.Bool"),
        ("syft.lib.python.List.__delattr__", "syft.lib.python.None"),
        ("syft.lib.python.List.__delitem__", "syft.lib.python.None"),
        ("syft.lib.python.List.__eq__", "syft.lib.python.Bool"),
        ("syft.lib.python.List.__mul__", "syft.lib.python.List"),
        ("syft.lib.python.List.__ne__", "syft.lib.python.Bool"),
        ("syft.lib.python.List.__sizeof__", "syft.lib.python.Int"),
        ("syft.lib.python.List.__len__", "syft.lib.python.Int"),
        ("syft.lib.python.List.__getitem__", "syft.lib.python.Any"),
        ("syft.lib.python.List.__setitem__", "syft.lib.python._SyNone"),
        ("syft.lib.python.List.__rmul__", "syft.lib.python.List"),
        ("syft.lib.python.List.copy", "syft.lib.python.List"),
        ("syft.lib.python.List.count", "syft.lib.python.Int"),
        ("syft.lib.python.List.sort", "syft.lib.python._SyNone"),
        ("syft.lib.python.List.reverse", "syft.lib.python._SyNone"),
        ("syft.lib.python.List.remove", "syft.lib.python._SyNone"),
        ("syft.lib.python.List.pop", "syft.lib.python.Any"),
        ("syft.lib.python.List.index", "syft.lib.python.Any"),
        ("syft.lib.python.List.insert", "syft.lib.python._SyNone"),
        ("syft.lib.python.List.clear", "syft.lib.python._SyNone"),
        ("syft.lib.python.List.extend", "syft.lib.python._SyNone"),
        ("syft.lib.python.List.__reversed__", "syft.lib.python.Iterator"),
        ("syft.lib.python.List.__delitem__", "syft.lib.python._SyNone"),
        # Bool methods - quite there
        ("syft.lib.python.Bool.__abs__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__eq__", "syft.lib.python.Bool"),
        ("syft.lib.python.Bool.__add__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__and__", "syft.lib.python.Bool"),
        ("syft.lib.python.Bool.__ceil__", "syft.lib.python.Bool"),
        ("syft.lib.python.Bool.__divmod__", "syft.lib.python.Tuple"),
        ("syft.lib.python.Bool.__floor__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__float__", "syft.lib.python.Float"),
        ("syft.lib.python.Bool.__floordiv__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__ge__", "syft.lib.python.Bool"),
        ("syft.lib.python.Bool.__gt__", "syft.lib.python.Bool"),
        ("syft.lib.python.Bool.__invert__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__le__", "syft.lib.python.Bool"),
        ("syft.lib.python.Bool.__lt__", "syft.lib.python.Bool"),
        ("syft.lib.python.Bool.__lshift__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__mod__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__mul__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__ne__", "syft.lib.python.Bool"),
        ("syft.lib.python.Bool.__neg__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__or__", "syft.lib.python.Bool"),
        ("syft.lib.python.Bool.__pos__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__pow__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__radd__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__rand__", "syft.lib.python.Bool"),
        ("syft.lib.python.Bool.__rdivmod__", "syft.lib.python.Bool"),
        ("syft.lib.python.Bool.__rfloordiv__", "syft.lib.python.Bool"),
        ("syft.lib.python.Bool.__rlshift__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__rmod__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__rmul__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__ror__", "syft.lib.python.Bool"),
        ("syft.lib.python.Bool.__round__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__rpow__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__rrshift__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__rshift__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__rsub__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__rtruediv__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__rxor__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__sub__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__truediv__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__xor__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.__trunc__", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.conjugate", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.bit_length", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.as_integer_ratio", "syft.lib.python.Tuple"),
        ("syft.lib.python.Bool.numerator", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.real", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.imag", "syft.lib.python.Int"),
        ("syft.lib.python.Bool.denominator", "syft.lib.python.Int"),
        # Float methods - subject to further change due
        ("syft.lib.python.Float.__add__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__truediv__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__divmod__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__eq__", "syft.lib.python.Bool"),
        ("syft.lib.python.Float.__ge__", "syft.lib.python.Bool"),
        ("syft.lib.python.Float.__lt__", "syft.lib.python.Bool"),
        ("syft.lib.python.Float.__le__", "syft.lib.python.Bool"),
        ("syft.lib.python.Float.__gt__", "syft.lib.python.Bool"),
        ("syft.lib.python.Float.__add__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__abs__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__bool__", "syft.lib.python.Bool"),
        ("syft.lib.python.Float.__sub__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__rsub__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__mul__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__rmul__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__divmod__", "syft.lib.python.Tuple"),
        ("syft.lib.python.Float.__int__", "syft.lib.python.Int"),
        ("syft.lib.python.Float.__neg__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__ne__", "syft.lib.python.Bool"),
        ("syft.lib.python.Float.__floordiv__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__truediv__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__mod__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__rmod__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__rdivmod__", "syft.lib.python.Tuple"),
        ("syft.lib.python.Float.__rfloordiv__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__round__", "syft.lib.python.Int"),
        ("syft.lib.python.Float.__rtruediv__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__sizeof__", "syft.lib.python.Int"),
        ("syft.lib.python.Float.__trunc__", "syft.lib.python.Int"),
        ("syft.lib.python.Float.as_integer_ratio", "syft.lib.python.Tuple"),
        ("syft.lib.python.Float.is_integer", "syft.lib.python.Bool"),
        ("syft.lib.python.Float.__pow__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__rpow__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__iadd__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__isub__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__imul__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__imod__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__ipow__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.__pos__", "syft.lib.python.Float"),
        ("syft.lib.python.Float.conjugate", "syft.lib.python.Float"),
        ("syft.lib.python.Float.imag", "syft.lib.python.Int"),
        ("syft.lib.python.Float.real", "syft.lib.python.Float"),
        # String Methods
        ("syft.lib.python.String.__add__", "syft.lib.python.String"),
        ("syft.lib.python.String.__contains__", "syft.lib.python.Bool"),
        ("syft.lib.python.String.__eq__", "syft.lib.python.Bool"),
        ("syft.lib.python.String.__float__", "syft.lib.python.Float"),
        ("syft.lib.python.String.__ge__", "syft.lib.python.Bool"),
        ("syft.lib.python.String.__getitem__", "syft.lib.python.String"),
        ("syft.lib.python.String.__gt__", "syft.lib.python.Bool"),
        ("syft.lib.python.String.__int__", "syft.lib.python.Int"),
        ("syft.lib.python.String.__iter__", "syft.lib.python.Any"),
        ("syft.lib.python.String.__le__", "syft.lib.python.Bool"),
        ("syft.lib.python.String.__len__", "syft.lib.python.Int"),
        ("syft.lib.python.String.__lt__", "syft.lib.python.Bool"),
        ("syft.lib.python.String.__mod__", "syft.lib.python.String"),
        ("syft.lib.python.String.__mul__", "syft.lib.python.String"),
        ("syft.lib.python.String.__ne__", "syft.lib.python.Bool"),
        ("syft.lib.python.String.__reversed__", "syft.lib.python.String"),
        ("syft.lib.python.String.__sizeof__", "syft.lib.python.Int"),
        ("syft.lib.python.String.__str__", "syft.lib.python.String"),
        ("syft.lib.python.String.capitalize", "syft.lib.python.String"),
        ("syft.lib.python.String.casefold", "syft.lib.python.String"),
        ("syft.lib.python.String.center", "syft.lib.python.String"),
        ("syft.lib.python.String.count", "syft.lib.python.Int"),
        ("syft.lib.python.String.encode", "syft.lib.python.String"),
        ("syft.lib.python.String.expandtabs", "syft.lib.python.String"),
        ("syft.lib.python.String.find", "syft.lib.python.Int"),
        ("syft.lib.python.String.format", "syft.lib.python.String"),
        ("syft.lib.python.String.format_map", "syft.lib.python.String"),
        ("syft.lib.python.String.index", "syft.lib.python.Int"),
        ("syft.lib.python.String.isalnum", "syft.lib.python.Bool"),
        ("syft.lib.python.String.isalpha", "syft.lib.python.Bool"),
        ("syft.lib.python.String.isdecimal", "syft.lib.python.Bool"),
        ("syft.lib.python.String.isdigit", "syft.lib.python.Bool"),
        ("syft.lib.python.String.isidentifier", "syft.lib.python.Bool"),
        ("syft.lib.python.String.islower", "syft.lib.python.Bool"),
        ("syft.lib.python.String.isnumeric", "syft.lib.python.Bool"),
        ("syft.lib.python.String.isprintable", "syft.lib.python.Bool"),
        ("syft.lib.python.String.isspace", "syft.lib.python.Bool"),
        ("syft.lib.python.String.isupper", "syft.lib.python.Bool"),
        ("syft.lib.python.String.join", "syft.lib.python.String"),
        ("syft.lib.python.String.ljust", "syft.lib.python.String"),
        ("syft.lib.python.String.lower", "syft.lib.python.String"),
        ("syft.lib.python.String.lstrip", "syft.lib.python.String"),
        ("syft.lib.python.String.partition", "syft.lib.python.Tuple"),
        ("syft.lib.python.String.replace", "syft.lib.python.String"),
        ("syft.lib.python.String.rfind", "syft.lib.python.Int"),
        ("syft.lib.python.String.rindex", "syft.lib.python.Int"),
        ("syft.lib.python.String.rjust", "syft.lib.python.String"),
        ("syft.lib.python.String.rpartition", "syft.lib.python.Tuple"),
        ("syft.lib.python.String.rsplit", "syft.lib.python.List"),
        ("syft.lib.python.String.rstrip", "syft.lib.python.String"),
        ("syft.lib.python.String.split", "syft.lib.python.List"),
        ("syft.lib.python.String.splitlines", "syft.lib.python.List"),
        ("syft.lib.python.String.startswith", "syft.lib.python.Bool"),
        ("syft.lib.python.String.strip", "syft.lib.python.String"),
        ("syft.lib.python.String.swapcase", "syft.lib.python.String"),
        ("syft.lib.python.String.title", "syft.lib.python.String"),
        ("syft.lib.python.String.translate", "syft.lib.python.String"),
        ("syft.lib.python.String.upper", "syft.lib.python.String"),
        ("syft.lib.python.String.zfill", "syft.lib.python.String"),
        ("syft.lib.python.String.__contains__", "syft.lib.python.Bool"),
        ("syft.lib.python.String.__rmul__", "syft.lib.python.String"),
        ("syft.lib.python.String.endswith", "syft.lib.python.Bool"),
        ("syft.lib.python.String.isascii", "syft.lib.python.Bool"),
        ("syft.lib.python.String.istitle", "syft.lib.python.Bool"),
        # Dict methods
        ("syft.lib.python.Dict.__contains__", "syft.lib.python.Bool"),
        ("syft.lib.python.Dict.__eq__", "syft.lib.python.Bool"),
        ("syft.lib.python.Dict.__format__", "syft.lib.python.String"),
        ("syft.lib.python.Dict.__ge__", "syft.lib.python.Bool"),
        ("syft.lib.python.Dict.__getitem__", "syft.lib.python.Any"),
        ("syft.lib.python.Dict.__gt__", "syft.lib.python.Bool"),
        ("syft.lib.python.Dict.__iter__", "syft.lib.python.Iterator"),
        ("syft.lib.python.Dict.__le__", "syft.lib.python.Bool"),
        ("syft.lib.python.Dict.__len__", "syft.lib.python.Int"),
        ("syft.lib.python.Dict.__lt__", "syft.lib.python.Bool"),
        ("syft.lib.python.Dict.__ne__", "syft.lib.python.Bool"),
        ("syft.lib.python.Dict.__sizeof__", "syft.lib.python.Int"),
        ("syft.lib.python.Dict.__str__", "syft.lib.python.String"),
        ("syft.lib.python.Dict.copy", "syft.lib.python.Dict"),
        ("syft.lib.python.Dict.fromkeys", "syft.lib.python.Dict"),
        # Rename get to dict_get because of conflict
        ("syft.lib.python.Dict.dict_get", "syft.lib.python.Any"),
        ("syft.lib.python.Dict.items", "syft.lib.python.Iterator"),
        ("syft.lib.python.Dict.keys", "syft.lib.python.Iterator"),
        ("syft.lib.python.Dict.pop", "syft.lib.python.Any"),
        ("syft.lib.python.Dict.popitem", "syft.lib.python.Tuple"),
        ("syft.lib.python.Dict.setdefault", "syft.lib.python.Any"),
        ("syft.lib.python.Dict.values", "syft.lib.python.Iterator"),
        # Int methods - subject to further change
        ("syft.lib.python.Int.__add__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__truediv__", "syft.lib.python.Float"),
        ("syft.lib.python.Int.__divmod__", "syft.lib.python.Float"),
        ("syft.lib.python.Int.__floordiv__", "syft.lib.python.Float"),
        ("syft.lib.python.Int.__invert__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__abs__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__bool__", "syft.lib.python.Bool"),
        ("syft.lib.python.Int.__divmod__", "syft.lib.python.Tuple"),
        ("syft.lib.python.Int.__rdivmod__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__radd__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__sub__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__rsub__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__rtruediv__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__mul__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__rmul__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__ceil__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__eq__", "syft.lib.python.Bool"),
        ("syft.lib.python.Int.__float__", "syft.lib.python.Float"),
        ("syft.lib.python.Int.__floor__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__floordiv__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__rfloordiv__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__truediv__", "syft.lib.python.Float"),
        ("syft.lib.python.Int.__mod__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__rmod__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__pow__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__rpow__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__lshift__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__rlshift__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__round__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__rshift__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__rrshift__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__and__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__rand__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__xor__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__xor__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__rxor__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__or__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__ror__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__ge__", "syft.lib.python.Bool"),
        ("syft.lib.python.Int.__lt__", "syft.lib.python.Bool"),
        ("syft.lib.python.Int.__le__", "syft.lib.python.Bool"),
        ("syft.lib.python.Int.__gt__", "syft.lib.python.Bool"),
        ("syft.lib.python.Int.__iadd__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__isub__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__imul__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__ifloordiv__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__itruediv__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__imod__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__ipow__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__ne__", "syft.lib.python.Bool"),
        ("syft.lib.python.Int.__neg__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__pos__", "syft.lib.python.Int"),
        ("syft.lib.python.Int.as_integer_ratio", "syft.lib.python.Tuple"),
        ("syft.lib.python.Int.bit_length", "syft.lib.python.Int"),
        ("syft.lib.python.Int.denominator", "syft.lib.python.Int"),
        ("syft.lib.python.Int.from_bytes", "syft.lib.python.Int"),
        ("syft.lib.python.Int.real", "syft.lib.python.Int"),
        ("syft.lib.python.Int.imag", "syft.lib.python.Int"),
        ("syft.lib.python.Int.numerator", "syft.lib.python.Int"),
        ("syft.lib.python.Int.conjugate", "syft.lib.python.Int"),
        ("syft.lib.python.Int.__trunc__", "syft.lib.python.Int"),
        # Tuple
        ("syft.lib.python.Tuple.__add__", "syft.lib.python.Tuple"),
        ("syft.lib.python.Tuple.__contains__", "syft.lib.python.Bool"),
        ("syft.lib.python.Tuple.__ne__", "syft.lib.python.Bool"),
        ("syft.lib.python.Tuple.__ge__", "syft.lib.python.Bool"),
        ("syft.lib.python.Tuple.__gt__", "syft.lib.python.Bool"),
        ("syft.lib.python.Tuple.__le__", "syft.lib.python.Bool"),
        ("syft.lib.python.Tuple.__lt__", "syft.lib.python.Bool"),
        ("syft.lib.python.Tuple.__mul__", "syft.lib.python.Tuple"),
        ("syft.lib.python.Tuple.__rmul__", "syft.lib.python.Tuple"),
        ("syft.lib.python.Tuple.__len__", "syft.lib.python.Int"),
        ("syft.lib.python.Tuple.__getitem__", "syft.lib.python.Any"),
        ("syft.lib.python.Tuple.count", "syft.lib.python.Int"),
        ("syft.lib.python.Tuple.index", "syft.lib.python.Int"),
        ("syft.lib.python.Tuple.__iter__", "syft.lib.python.Iterator"),
        # PyContainer - quite there
        ("syft.lib.python.Any.__add__", "syft.lib.python.Any"),
        ("syft.lib.python.Any.__iter__", "syft.lib.python.Iterator"),
        ("syft.lib.python.Any.__next__", "syft.lib.python.Any"),
        ("syft.lib.python.Any.__len__", "syft.lib.python.Int"),
        ("syft.lib.python.Any.__radd__", "syft.lib.python.Any"),
        ("syft.lib.python.Any.__truediv__", "syft.lib.python.Any"),
        ("syft.lib.python.Any.__rtruediv__", "syft.lib.python.Any"),
        ("syft.lib.python.Any.__floordiv__", "syft.lib.python.Any"),
        ("syft.lib.python.Any.__rfloordiv__", "syft.lib.python.Any"),
        ("syft.lib.python.Any.__mul__", "syft.lib.python.Any"),
        ("syft.lib.python.Any.__rmul__", "syft.lib.python.Any"),
        ("syft.lib.python.Any.__sub__", "syft.lib.python.Any"),
        ("syft.lib.python.Any.__rsub__", "syft.lib.python.Any"),
        (
            "syft.lib.python.Iterator.__next__",
            UnionGenerator[
                "syft.lib.python.Int",
                "syft.lib.python.Float",
                "syft.lib.python.String",
                "torch.nn.Parameter",
                "torch.Tensor",
            ],
        ),  # temp until casting
        ("syft.lib.python.Iterator.__iter__", "syft.lib.python.Iterator"),
        ("syft.lib.python.Iterator.__len__", "syft.lib.python.Int"),
        ("syft.lib.python.Set.__and__", "syft.lib.python.Set"),
        ("syft.lib.python.Set.__contains__", "syft.lib.python.Bool"),
        ("syft.lib.python.Set.__eq__", "syft.lib.python.Bool"),
        ("syft.lib.python.Set.__ge__", "syft.lib.python.Bool"),
        ("syft.lib.python.Set.__gt__", "syft.lib.python.Bool"),
        ("syft.lib.python.Set.__iand__", "syft.lib.python.Set"),
        ("syft.lib.python.Set.__ior__", "syft.lib.python.Set"),
        ("syft.lib.python.Set.__isub__", "syft.lib.python.Set"),
        ("syft.lib.python.Set.__ixor__", "syft.lib.python.Set"),
        ("syft.lib.python.Set.__le__", "syft.lib.python.Bool"),
        ("syft.lib.python.Set.__len__", "syft.lib.python.Int"),
        ("syft.lib.python.Set.__lt__", "syft.lib.python.Bool"),
        ("syft.lib.python.Set.__ne__", "syft.lib.python.Bool"),
        ("syft.lib.python.Set.__or__", "syft.lib.python.Set"),
        ("syft.lib.python.Set.__sub__", "syft.lib.python.Set"),
        ("syft.lib.python.Set.__xor__", "syft.lib.python.Set"),
        ("syft.lib.python.Set.add", "syft.lib.python._SyNone"),
        ("syft.lib.python.Set.clear", "syft.lib.python._SyNone"),
        ("syft.lib.python.Set.difference", "syft.lib.python.Set"),
        ("syft.lib.python.Set.difference_update", "syft.lib.python._SyNone"),
        ("syft.lib.python.Set.discard", "syft.lib.python._SyNone"),
        ("syft.lib.python.Set.intersection", "syft.lib.python.Set"),
        ("syft.lib.python.Set.intersection_update", "syft.lib.python._SyNone"),
        ("syft.lib.python.Set.isdisjoint", "syft.lib.python.Bool"),
        ("syft.lib.python.Set.issuperset", "syft.lib.python.Bool"),
        ("syft.lib.python.Set.pop", "syft.lib.python._SyNone"),
        ("syft.lib.python.Set.remove", "syft.lib.python._SyNone"),
        (
            "syft.lib.python.Set.symmetric_difference_update",
            "syft.lib.python._SyNone",
        ),
        ("syft.lib.python.Set.symmetric_difference", "syft.lib.python.Set"),
        ("syft.lib.python.Set.union", "syft.lib.python.Set"),
        ("syft.lib.python.Set.update", "syft.lib.python._SyNone"),
        (
            "syft.lib.python.collections.OrderedDict.__contains__",
            "syft.lib.python.Bool",
        ),
        (
            "syft.lib.python.collections.OrderedDict.__delitem__",
            "syft.lib.python._SyNone",
        ),
        (
            "syft.lib.python.collections.OrderedDict.__eq__",
            "syft.lib.python.Bool",
        ),
        (
            "syft.lib.python.collections.OrderedDict.__ge__",
            "syft.lib.python.Bool",
        ),
        (
            "syft.lib.python.collections.OrderedDict.__getitem__",
            "syft.lib.python.Any",
        ),
        (
            "syft.lib.python.collections.OrderedDict.__gt__",
            "syft.lib.python.Bool",
        ),
        (
            "syft.lib.python.collections.OrderedDict.__le__",
            "syft.lib.python.Bool",
        ),
        (
            "syft.lib.python.collections.OrderedDict.__iter__",
            "syft.lib.python.Iterator",
        ),
        ("syft.lib.python.collections.OrderedDict.__len__", "syft.lib.python.Int"),
        (
            "syft.lib.python.collections.OrderedDict.__lt__",
            "syft.lib.python.Bool",
        ),
        (
            "syft.lib.python.collections.OrderedDict.__ne__",
            "syft.lib.python.Bool",
        ),
        (
            "syft.lib.python.collections.OrderedDict.__setitem__",
            "syft.lib.python._SyNone",
        ),
        (
            "syft.lib.python.collections.OrderedDict.clear",
            "syft.lib.python._SyNone",
        ),
        (
            "syft.lib.python.collections.OrderedDict.copy",
            "syft.lib.python.collections.OrderedDict",
        ),
        (
            "syft.lib.python.collections.OrderedDict.fromkeys",
            "syft.lib.python.collections.OrderedDict",
        ),
        ("syft.lib.python.collections.OrderedDict.items", "syft.lib.python.Iterator"),
        ("syft.lib.python.collections.OrderedDict.keys", "syft.lib.python.Iterator"),
        (
            "syft.lib.python.collections.OrderedDict.move_to_end",
            "syft.lib.python._SyNone",
        ),
        ("syft.lib.python.collections.OrderedDict.pop", "syft.lib.python.Any"),
        ("syft.lib.python.collections.OrderedDict.popitem", "syft.lib.python.Any"),
        (
            "syft.lib.python.collections.OrderedDict.setdefault",
            "syft.lib.python.Any",
        ),
        (
            "syft.lib.python.collections.OrderedDict.update",
            "syft.lib.python._SyNone",
        ),
        (
            "syft.lib.python.collections.OrderedDict.values",
            "syft.lib.python.Iterator",
        ),
        ("syft.lib.python.collections.OrderedDict.items", "syft.lib.python.List"),
        (
            "syft.lib.python.collections.OrderedDict.dict_get",
            "syft.lib.python.Any",
        ),
    ]

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast



# stdlib
from typing import Any
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft relative
from ... import deserialize
from ... import serialize
from ...core.common import UID
from ...core.common.serde.serializable import bind_protobuf
from ...logger import traceback_and_raise
from ...proto.lib.python.complex_pb2 import Complex as Complex_PB
from .primitive_factory import PrimitiveFactory
from .primitive_interface import PyPrimitive
from .types import SyPrimitiveRet


# TODO - actually make all of this work
@bind_protobuf
class Complex(complex, PyPrimitive):
    def __new__(
        self, real: Any = None, imag: Any = None, id: Optional[UID] = None
    ) -> "Complex":
        if real is None:
            return complex.__new__(self)
        if imag is None:
            return complex.__new__(self, real=real)
        if isinstance(real, str):
            traceback_and_raise(
                TypeError("Complex() can't take second arg if first is a string")
            )
        return complex.__new__(self, real=real, imag=imag)

    def __init__(self, real: Any = None, imag: Any = None, id: Optional[UID] = None):
        complex.__init__(self)
        self._id = id or UID()

    @property
    def id(self) -> UID:
        """We reveal PyPrimitive.id as a property to discourage users and
        developers of Syft from modifying .id attributes after an object
        has been initialized.

        :return: returns the unique id of the object
        :rtype: UID
        """
        return self._id

    def upcast(self) -> complex:
        return complex(self)

    def __add__(self, x: complex) -> SyPrimitiveRet:
        result = complex.__add__(self, x)
        return PrimitiveFactory.generate_primitive(value=result)

    def __sub__(self, x: complex) -> SyPrimitiveRet:
        result = complex.__sub__(self, x)
        return PrimitiveFactory.generate_primitive(value=result)

    def __mul__(self, x: complex) -> SyPrimitiveRet:
        result = complex.__mul__(self, x)
        return PrimitiveFactory.generate_primitive(value=result)

    def __pow__(self, x: complex, modulo: Any = 10) -> SyPrimitiveRet:
        result = complex.__pow__(self, x, modulo)
        return PrimitiveFactory.generate_primitive(value=result)

    def __truediv__(self, x: complex) -> SyPrimitiveRet:
        result = complex.__truediv__(self, x)
        return PrimitiveFactory.generate_primitive(value=result)

    def __radd__(self, x: complex) -> SyPrimitiveRet:
        result = complex.__radd__(self, x)
        return PrimitiveFactory.generate_primitive(value=result)

    def __rsub__(self, x: complex) -> SyPrimitiveRet:
        result = complex.__rsub__(self, x)
        return PrimitiveFactory.generate_primitive(value=result)

    def __rmul__(self, x: complex) -> SyPrimitiveRet:
        result = complex.__rmul__(self, x)
        return PrimitiveFactory.generate_primitive(value=result)

    def __rpow__(self, x: complex, modulo: Optional[Any] = None) -> SyPrimitiveRet:
        if modulo:
            return PrimitiveFactory.generate_primitive(
                value=complex.__rpow__(self, x, modulo)
            )
        return PrimitiveFactory.generate_primitive(value=complex.__rpow__(self, x))

    def __rtruediv__(self, x: complex) -> SyPrimitiveRet:
        result = complex.__rtruediv__(self, x)
        return PrimitiveFactory.generate_primitive(value=result)

    def __eq__(self, x: object) -> SyPrimitiveRet:
        result = complex.__eq__(self, x)
        return PrimitiveFactory.generate_primitive(value=result)

    def __ne__(self, x: object) -> SyPrimitiveRet:
        result = complex.__ne__(self, x)
        return PrimitiveFactory.generate_primitive(value=result)

    def __neg__(self) -> SyPrimitiveRet:
        result = complex.__neg__(self)
        return PrimitiveFactory.generate_primitive(value=result)

    def __pos__(self) -> SyPrimitiveRet:
        result = complex.__pos__(self)
        return PrimitiveFactory.generate_primitive(value=result)

    # def __str__(self) -> PyPrimitive:
    #     ...

    # def __complex__(self) -> "Complex":
    #     result = complex.__complex__()
    #     return PrimitiveFactory.generate_primitive(value=result)

    def __abs__(self) -> SyPrimitiveRet:
        result = complex.__abs__(self)
        return PrimitiveFactory.generate_primitive(value=result)

    def __hash__(self) -> SyPrimitiveRet:
        result = complex.__hash__(self)
        return PrimitiveFactory.generate_primitive(value=result)

    def __bool__(self) -> bool:
        # NOTE we return a real bool here, not a syft Bool
        return complex.__bool__(self)

    def _object2proto(self) -> Complex_PB:
        return Complex_PB(id=serialize(obj=self.id), real=self.real, imag=self.imag)

    @staticmethod
    def _proto2object(proto: Complex_PB) -> "Complex":
        return Complex(
            id=deserialize(blob=proto.id),
            real=proto.real,
            imag=proto.imag,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return Complex_PB



# stdlib
from typing import Any
from typing import Iterator
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
import pandas as pd
import torch as th
from torch.utils.data import DataLoader
from torch.utils.data import Dataset

# syft relative
from ... import deserialize
from ... import serialize
from ...logger import traceback_and_raise
from ...proto.core.remote_dataloader.remote_dataset_pb2 import (
    RemoteDataLoader as RemoteDataLoader_PB,
)
from ...proto.core.remote_dataloader.remote_dataset_pb2 import (
    RemoteDataset as RemoteDataset_PB,
)
from ..common.serde.serializable import Serializable
from ..common.serde.serializable import bind_protobuf

DATA_TYPE_TORCH_TENSOR = "torch_tensor"
DATA_TYPE_PANDAS_CSV = "pandas_csv"


@bind_protobuf
class RemoteDataset(Dataset, Serializable):
    def __init__(self, path: str, data_type: str = DATA_TYPE_TORCH_TENSOR):
        """
        Arguments:
            path: information about where to get the raw data, for example, a file path,
            or a directory path data_type: the type of data for example torch_tensor
            or pandas_csv
        For now, it's should simply be a .pt / .csv file, which stores a Dataset object.
        """
        self.path = path
        self.data_type = data_type

    def load_dataset(self) -> None:
        """
        Load the real Dataset object on DO's machine.
        But for now, it's just simply calling torch.load on a .pt file.
        """
        if self.data_type == DATA_TYPE_TORCH_TENSOR:
            self.dataset = th.load(self.path)
        elif self.data_type == DATA_TYPE_PANDAS_CSV:
            self.dataset = pd.read_csv(self.path)

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(
        self, key: Union[str, int, slice, pd.DataFrame, pd.Series]
    ) -> Union[th.Tensor, pd.DataFrame, pd.Series]:
        return self.dataset[key]

    def __repr__(self) -> str:
        return f"{type(self)}: {self.data_type}"

    def _object2proto(self) -> RemoteDataset_PB:
        proto = RemoteDataset_PB()
        proto.path = self.path
        proto.data_type = self.data_type
        return proto

    @staticmethod
    def _proto2object(proto: Any) -> "RemoteDataset":
        return RemoteDataset(path=proto.path, data_type=proto.data_type)

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return RemoteDataset_PB


@bind_protobuf
class RemoteDataLoader(Serializable):
    def __init__(self, remote_dataset: RemoteDataset, batch_size: int = 1):
        """
        TODO: now, only batch_size can be passed in by users, and it's used when create
        DataLoader object in self.create_dataloader. In future steps, more augmentations
        should be supported, like shuffle, sampler, collate_fn, etc.
        """
        self.remote_dataset = remote_dataset
        self.batch_size = batch_size

    def _object2proto(self) -> RemoteDataLoader_PB:
        proto = RemoteDataLoader_PB()
        proto.batch_size = self.batch_size
        proto.remote_dataset.CopyFrom(serialize(self.remote_dataset))
        return proto

    @staticmethod
    def _proto2object(proto: Any) -> "RemoteDataLoader":
        remote_dataset = deserialize(proto.remote_dataset)
        batch_size = proto.batch_size
        return RemoteDataLoader(remote_dataset=remote_dataset, batch_size=batch_size)

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return RemoteDataLoader_PB

    def load_dataset(self) -> None:
        self.remote_dataset.load_dataset()

    def create_dataloader(self) -> None:
        if self.remote_dataset.data_type == DATA_TYPE_TORCH_TENSOR:
            dataset = getattr(self.remote_dataset, "dataset", None)
            self.dataloader = DataLoader(dataset=dataset, batch_size=self.batch_size)
        else:
            traceback_and_raise(
                ValueError(
                    "Cannot create a DataLoader for type: {self.remote_dataset.data_type}"
                )
            )

    def __len__(self) -> int:
        return len(self.dataloader)

    def __iter__(self) -> "Iterator":
        return iter(self.dataloader)



# stdlib
from typing import Any

# third party
import pytest

# syft absolute
import syft as sy

# syft relative
from .utils_test import decrypt

ts = pytest.importorskip("tenseal")
sy.load("tenseal")


@pytest.fixture(scope="function")
def context() -> Any:
    context = ts.context(ts.SCHEME_TYPE.BFV, 8192, 1032193, n_threads=1)
    return context


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_bfvvector_sanity(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    v1 = [0, 1, 2, 3, 4]
    enc_v1 = ts.bfv_vector(context, v1)

    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)

    result = decrypt(context, enc_v1_ptr)
    assert result == [0, 1, 2, 3, 4]


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_bfvvector_add(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    v1 = [0, 1, 2, 3, 4]
    v2 = [4, 3, 2, 1, 0]
    expected = [v1 + v2 for v1, v2 in zip(v1, v2)]

    enc_v1 = ts.bfv_vector(context, v1)
    enc_v2 = ts.bfv_vector(context, v2)

    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)
    enc_v2_ptr = enc_v2.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)
    enc_v2_ptr.link_context(ctx_ptr)

    # add
    result_enc_ptr = enc_v1_ptr + enc_v2_ptr

    result = decrypt(context, result_enc_ptr)
    assert result == expected

    # add inplace
    enc_v1_ptr += enc_v2_ptr

    result = decrypt(context, enc_v1_ptr)
    assert result == expected


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_bfvvector_sub(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    v1 = [0, 1, 2, 3, 4]
    v2 = [4, 3, 2, 1, 0]
    expected = [v1 - v2 for v1, v2 in zip(v1, v2)]

    enc_v1 = ts.bfv_vector(context, v1)
    enc_v2 = ts.bfv_vector(context, v2)

    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)
    enc_v2_ptr = enc_v2.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)
    enc_v2_ptr.link_context(ctx_ptr)

    # sub
    result_enc_ptr = enc_v1_ptr - enc_v2_ptr

    result = decrypt(context, result_enc_ptr)
    assert result == expected

    # sub inplace
    enc_v1_ptr -= enc_v2_ptr

    result = decrypt(context, enc_v1_ptr)
    assert result == expected


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_bfvvector_mul(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    v1 = [0, 1, 2, 3, 4]
    v2 = [4, 3, 2, 1, 0]
    expected = [v1 * v2 for v1, v2 in zip(v1, v2)]

    enc_v1 = ts.bfv_vector(context, v1)
    enc_v2 = ts.bfv_vector(context, v2)

    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)
    enc_v2_ptr = enc_v2.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)
    enc_v2_ptr.link_context(ctx_ptr)

    # mul
    result_enc_ptr = enc_v1_ptr * enc_v2_ptr

    result = decrypt(context, result_enc_ptr)
    assert result == expected

    # mul inplace
    enc_v1_ptr *= enc_v2_ptr

    result = decrypt(context, enc_v1_ptr)
    assert result == expected


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_bfvvector_iadd(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    v1 = [0, 1, 2, 3, 4]
    v2 = [4, 3, 2, 1, 0]
    expected = [v1 + v2 for v1, v2 in zip(v1, v2)]

    enc_v1 = ts.bfv_vector(context, v1)

    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)

    # iadd
    result_enc_ptr = enc_v1_ptr + v2

    result = decrypt(context, result_enc_ptr)
    assert result == expected

    # radd
    result_enc_ptr = v2 + enc_v1_ptr

    result = decrypt(context, result_enc_ptr)
    assert result == expected

    # iadd inplace
    enc_v1_ptr += v2

    result = decrypt(context, enc_v1_ptr)
    assert result == expected


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_bfvvector_isub(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    v1 = [0, 1, 2, 3, 4]
    v2 = [4, 3, 2, 1, 0]
    expected = [v1 - v2 for v1, v2 in zip(v1, v2)]

    enc_v1 = ts.bfv_vector(context, v1)

    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)

    # isub
    result_enc_ptr = enc_v1_ptr - v2

    result = decrypt(context, result_enc_ptr)
    assert result == expected


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_bfvvector_imul(
    context: Any, root_client: sy.VirtualMachineClient
) -> None:
    v1 = [0, 1, 2, 3, 4]
    v2 = [4, 3, 2, 1, 0]
    expected = [v1 * v2 for v1, v2 in zip(v1, v2)]

    enc_v1 = ts.bfv_vector(context, v1)

    ctx_ptr = context.send(root_client, pointable=True)
    enc_v1_ptr = enc_v1.send(root_client, pointable=True)

    enc_v1_ptr.link_context(ctx_ptr)

    # imul
    result_enc_ptr = enc_v1_ptr * v2

    result = decrypt(context, result_enc_ptr)
    assert result == expected

    # rmul
    result_enc_ptr = v2 * enc_v1_ptr

    result = decrypt(context, result_enc_ptr)
    assert result == expected



# stdlib
from typing import Any
from typing import Optional

# third party
from torch import device

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...proto.lib.torch.device_pb2 import Device as Device_PB

# use -2 to represent index=None
INDEX_NONE = -2


def object2proto(obj: device) -> "Device_PB":
    proto = Device_PB()
    proto.type = obj.type
    proto.index = INDEX_NONE if obj.index is None else obj.index
    return proto


def proto2object(proto: "Device_PB") -> Any:
    device_type = proto.type
    index: Optional[int] = None if proto.index == INDEX_NONE else proto.index
    obj = device(device_type, index)
    return obj


GenerateWrapper(
    wrapped_type=device,
    import_path="torch.device",
    protobuf_scheme=Device_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



# future
from __future__ import print_function

# stdlib
import argparse

# third party
import matplotlib
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim

matplotlib.use("Agg")
# third party
import matplotlib.pyplot as plt


class Sequence(nn.Module):
    def __init__(self):
        super(Sequence, self).__init__()
        self.lstm1 = nn.LSTMCell(1, 51)
        self.lstm2 = nn.LSTMCell(51, 51)
        self.linear = nn.Linear(51, 1)

    def forward(self, input, future=0):
        outputs = []
        h_t = torch.zeros(input.size(0), 51, dtype=torch.double)
        c_t = torch.zeros(input.size(0), 51, dtype=torch.double)
        h_t2 = torch.zeros(input.size(0), 51, dtype=torch.double)
        c_t2 = torch.zeros(input.size(0), 51, dtype=torch.double)

        for input_t in input.split(1, dim=1):
            h_t, c_t = self.lstm1(input_t, (h_t, c_t))
            h_t2, c_t2 = self.lstm2(h_t, (h_t2, c_t2))
            output = self.linear(h_t2)
            outputs += [output]
        for i in range(future):  # if we should predict the future
            h_t, c_t = self.lstm1(output, (h_t, c_t))
            h_t2, c_t2 = self.lstm2(h_t, (h_t2, c_t2))
            output = self.linear(h_t2)
            outputs += [output]
        outputs = torch.cat(outputs, dim=1)
        return outputs


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--steps",
        type=int,
        default=15,
        help="steps to run; default:15"
    )
    opt = parser.parse_args()

    # set random seed to 0
    np.random.seed(0)
    torch.manual_seed(0)

    # load data and make training set
    data = torch.load("traindata.pt")
    input = torch.from_numpy(data[3:, :-1])
    target = torch.from_numpy(data[3:, 1:])
    test_input = torch.from_numpy(data[:3, :-1])
    test_target = torch.from_numpy(data[:3, 1:])

    # build the model
    seq = Sequence()
    seq.double()
    criterion = nn.MSELoss()

    # use LBFGS as optimizer since we can load the whole data to train
    optimizer = optim.LBFGS(seq.parameters(), lr=0.8)

    # begin to train
    for i in range(opt.steps):
        print("STEP: ", i)

        def closure():
            optimizer.zero_grad()
            out = seq(input)
            loss = criterion(out, target)
            print("loss:", loss.item())
            loss.backward()
            return loss

        optimizer.step(closure)

        # begin to predict, no need to track gradient here
        with torch.no_grad():
            future = 1000
            pred = seq(test_input, future=future)
            loss = criterion(pred[:, :-future], test_target)
            print("test loss:", loss.item())
            y = pred.detach().numpy()

        # draw the result
        plt.figure(figsize=(30, 10))
        plt.title(
            "Predict future values for time sequences\n(Dashlines are predicted values)",
            fontsize=30,
        )
        plt.xlabel("x", fontsize=20)
        plt.ylabel("y", fontsize=20)
        plt.xticks(fontsize=20)
        plt.yticks(fontsize=20)

        def draw(yi, color):
            plt.plot(
                np.arange(input.size(1)), yi[: input.size(1)], color, linewidth=2.0
            )
            plt.plot(
                np.arange(input.size(1), input.size(1) + future),
                yi[input.size(1) :],
                color + ":",
                linewidth=2.0,
            )

        draw(y[0], "r")
        draw(y[1], "g")
        draw(y[2], "b")
        plt.savefig("predict%d.pdf" % i)
        plt.close()



# stdlib
from typing import Any
from typing import Dict

# syft absolute
import syft as sy
from syft import serialize
from syft.core.io.address import Address
from syft.grid.messages.setup_messages import CreateInitialSetUpMessage
from syft.grid.messages.setup_messages import CreateInitialSetUpResponse
from syft.grid.messages.setup_messages import GetSetUpMessage
from syft.grid.messages.setup_messages import GetSetUpResponse


def test_create_initial_setup_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    request_content = {
        "settings": {
            "cloud-admin-token": "d84we35ad3a1d59a84sd9",
            "cloud-credentials": "<cloud-credentials.pem>",
            "infra": {"autoscaling": True, "triggers": {"memory": "50", "vCPU": "80"}},
        }
    }
    msg = CreateInitialSetUpMessage(
        address=target,
        content=request_content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_create_initial_setup_response_serde() -> None:
    target = Address(name="Alice")

    request_content = {"msg": "Initial setup registered successfully!"}
    msg = CreateInitialSetUpResponse(
        address=target,
        status_code=200,
        content=request_content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_initial_setup_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    request_content: Dict[Any, Any] = {}
    msg = GetSetUpMessage(
        address=target,
        content=request_content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_delete_worker_response_serde() -> None:
    target = Address(name="Alice")

    content = {
        "settings": {
            "cloud-admin-token": "d84we35ad3a1d59a84sd9",
            "cloud-credentials": "<cloud-credentials.pem>",
            "infra": {"autoscaling": True, "triggers": {"memory": "50", "vCPU": "80"}},
        }
    }
    msg = GetSetUpResponse(
        status_code=200,
        address=target,
        content=content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2



# stdlib
from typing import Any
from typing import List


def listify(x: Any) -> List[Any]:
    """turns x into a list.
    If x is a list or tuple, return as list.
    if x is not a list: return [x]
    if x is None: return []

    Args:
        x (Any): some object

    Returns:
        List[Any]: x, as a list
    """
    return list(x) if isinstance(x, (list, tuple)) else ([] if x is None else [x])



# stdlib
from typing import Any
from typing import Callable
from typing import Dict

# syft relative
from ...messages.association_messages import DeleteAssociationRequestMessage
from ...messages.association_messages import GetAssociationRequestMessage
from ...messages.association_messages import GetAssociationRequestsMessage
from ...messages.association_messages import RespondAssociationRequestMessage
from ...messages.association_messages import SendAssociationRequestMessage
from .request_api import GridRequestAPI


class AssociationRequestAPI(GridRequestAPI):
    response_key = "association-request"
    _accept = "accept"
    _deny = "deny"

    def __init__(self, send: Callable):
        super().__init__(
            create_msg=SendAssociationRequestMessage,
            get_msg=GetAssociationRequestMessage,
            get_all_msg=GetAssociationRequestsMessage,
            delete_msg=DeleteAssociationRequestMessage,
            send=send,
            response_key=AssociationRequestAPI.response_key,
        )

    def update(self, **kwargs: Any) -> Dict[Any, Any]:
        raise NotImplementedError("Method not implemented to Association Requests!")

    def __getitem__(self, key: int) -> Any:
        return self.get(association_request_id=key)

    def __delitem__(self, key: int) -> None:
        self.delete(association_request_id=key)

    def to_obj(self, result: Dict[Any, Any]) -> Any:
        _association_obj = super().to_obj(result)

        _content = {
            "address": _association_obj.address,
            "handshake": _association_obj.handshake_value,
            "sender_address": _association_obj.sender_address,
        }

        def _accept() -> Dict[str, str]:
            _content["value"] = AssociationRequestAPI._accept
            return self.send(
                grid_msg=RespondAssociationRequestMessage, content=_content
            )

        def _deny() -> Dict[str, str]:
            _content["value"] = AssociationRequestAPI._deny
            return self.send(
                grid_msg=RespondAssociationRequestMessage, content=_content
            )

        _association_obj.accept = _accept
        _association_obj.deny = _deny

        return _association_obj



# stdlib
import os
from pathlib import Path
from secrets import randbelow
from typing import Any
from typing import List
from typing import Optional
from typing import Union

# third party
from forbiddenfruit import curse
from nacl.signing import SigningKey
from nacl.signing import VerifyKey

# syft absolute
import syft

# syft relative
from .logger import critical
from .logger import debug
from .logger import error
from .logger import traceback_and_raise


def validate_type(_object: object, _type: type, optional: bool = False) -> Any:
    if isinstance(_object, _type) or (optional and (_object is None)):
        return _object

    traceback_and_raise(
        f"Object {_object} should've been of type {_type}, not {_object}."
    )


def validate_field(_object: object, _field: str) -> Any:
    object = getattr(_object, _field, None)

    if object is not None:
        return object

    traceback_and_raise(f"Object {_object} has no {_field} field set.")


def get_subclasses(obj_type: type) -> List[type]:
    """Recursively generate the list of all classes within the sub-tree of an object

    As a paradigm in Syft, we often allow for something to be known about by another
    part of the codebase merely because it has subclassed a particular object. While
    this can be a big "magicish" it also can simplify future extensions and reduce
    the likelihood of small mistakes (if done right).

    This is a utility function which allows us to look for sub-classes and the sub-classes
    of those sub-classes etc. returning a full list of descendants of a class in the inheritance
    hierarchy.

    Args:
        obj_type: the type we want to look for sub-classes of

    Returns:
        the list of subclasses of obj_type

    """

    classes = list()
    for sc in obj_type.__subclasses__():
        classes.append(sc)
        classes += get_subclasses(obj_type=sc)
    return classes


def index_modules(a_dict: object, keys: List[str]) -> object:
    """Recursively find a syft module from its path

    This is the recursive inner function of index_syft_by_module_name.
    See that method for a full description.

    Args:
        a_dict: a module we're traversing
        keys: the list of string attributes we're using to traverse the module

    Returns:
        a reference to the final object

    """

    if len(keys) == 0:
        return a_dict
    return index_modules(a_dict=a_dict.__dict__[keys[0]], keys=keys[1:])


def index_syft_by_module_name(fully_qualified_name: str) -> object:
    """Look up a Syft class/module/function from full path and name

    Sometimes we want to use the fully qualified name (such as one
    generated from the 'get_fully_qualified_name' method below) to
    fetch an actual reference. This is most commonly used in deserialization
    so that we can have generic protobuf objects which just have a string
    representation of the specific object it is meant to deserialize to.

    Args:
        fully_qualified_name: the name in str of a module, class, or function

    Returns:
        a reference to the actual object at that string path

    """
    attr_list = fully_qualified_name.split(".")

    # we deal with VerifyAll differently, because we don't it be imported and used by users
    if attr_list[-1] == "VerifyAll":
        return type(syft.core.common.group.VERIFYALL)

    if attr_list[0] != "syft":
        raise ReferenceError(f"Reference don't match: {attr_list[0]}")

    if (
        attr_list[1] != "core"
        and attr_list[1] != "lib"
        and attr_list[1] != "grid"
        and attr_list[1] != "wrappers"
    ):
        raise ReferenceError(f"Reference don't match: {attr_list[1]}")

    return index_modules(a_dict=globals()["syft"], keys=attr_list[1:])


def get_fully_qualified_name(obj: object) -> str:
    """Return the full path and name of a class

    Sometimes we want to return the entire path and name encoded
    using periods. For example syft.core.common.message.SyftMessage
    is the current fully qualified path and name for the SyftMessage
    object.

    Args:
        obj: the object we want to get the name of

    Returns:
        the full path and name of the object

    """
    fqn = obj.__module__
    try:
        fqn += "." + obj.__class__.__name__
    except Exception as e:
        error(f"Failed to get FQN: {e}")
    return fqn


def aggressive_set_attr(obj: object, name: str, attr: object) -> None:
    """Different objects prefer different types of monkeypatching - try them all

    Args:
        obj: object whose attribute has to be set
        name: attribute name
        attr: value given to the attribute

    """
    try:
        setattr(obj, name, attr)
    except Exception:
        curse(obj, name, attr)


def obj2pointer_type(obj: object) -> type:
    fqn = None
    try:
        fqn = get_fully_qualified_name(obj=obj)
    except Exception as e:
        # sometimes the object doesn't have a __module__ so you need to use the type
        # like: collections.OrderedDict
        debug(f"Unable to get get_fully_qualified_name of {type(obj)} trying type. {e}")
        if obj is None:
            fqn = "syft.lib.python._SyNone"
        else:
            fqn = get_fully_qualified_name(obj=type(obj))

    try:
        ref = syft.lib_ast.query(fqn, obj_type=type(obj))
    except Exception as e:
        log = f"Cannot find {type(obj)} {fqn} in lib_ast. {e}"
        critical(log)
        raise Exception(log)

    return ref.pointer_type


def key_emoji(key: object) -> str:
    try:
        if isinstance(key, (bytes, SigningKey, VerifyKey)):
            hex_chars = bytes(key).hex()[-8:]
            return char_emoji(hex_chars=hex_chars)
    except Exception as e:
        error(f"Fail to get key emoji: {e}")
        pass
    return "ALL"


def char_emoji(hex_chars: str) -> str:
    base = ord("\U0001F642")
    hex_base = ord("0")
    code = 0
    for char in hex_chars:
        offset = ord(char)
        code += offset - hex_base
    return chr(base + code)


left_name = [
    "admiring",
    "adoring",
    "affectionate",
    "agitated",
    "amazing",
    "angry",
    "awesome",
    "beautiful",
    "blissful",
    "bold",
    "boring",
    "brave",
    "busy",
    "charming",
    "clever",
    "cool",
    "compassionate",
    "competent",
    "condescending",
    "confident",
    "cranky",
    "crazy",
    "dazzling",
    "determined",
    "distracted",
    "dreamy",
    "eager",
    "ecstatic",
    "elastic",
    "elated",
    "elegant",
    "eloquent",
    "epic",
    "exciting",
    "fervent",
    "festive",
    "flamboyant",
    "focused",
    "friendly",
    "frosty",
    "funny",
    "gallant",
    "gifted",
    "goofy",
    "gracious",
    "great",
    "happy",
    "hardcore",
    "heuristic",
    "hopeful",
    "hungry",
    "infallible",
    "inspiring",
    "interesting",
    "intelligent",
    "jolly",
    "jovial",
    "keen",
    "kind",
    "laughing",
    "loving",
    "lucid",
    "magical",
    "mystifying",
    "modest",
    "musing",
    "naughty",
    "nervous",
    "nice",
    "nifty",
    "nostalgic",
    "objective",
    "optimistic",
    "peaceful",
    "pedantic",
    "pensive",
    "practical",
    "priceless",
    "quirky",
    "quizzical",
    "recursing",
    "relaxed",
    "reverent",
    "romantic",
    "sad",
    "serene",
    "sharp",
    "silly",
    "sleepy",
    "stoic",
    "strange",
    "stupefied",
    "suspicious",
    "sweet",
    "tender",
    "thirsty",
    "trusting",
    "unruffled",
    "upbeat",
    "vibrant",
    "vigilant",
    "vigorous",
    "wizardly",
    "wonderful",
    "xenodochial",
    "youthful",
    "zealous",
    "zen",
]

right_name = [
    "albattani",
    "allen",
    "almeida",
    "antonelli",
    "agnesi",
    "archimedes",
    "ardinghelli",
    "aryabhata",
    "austin",
    "babbage",
    "banach",
    "banzai",
    "bardeen",
    "bartik",
    "bassi",
    "beaver",
    "bell",
    "benz",
    "bhabha",
    "bhaskara",
    "black",
    "blackburn",
    "blackwell",
    "bohr",
    "booth",
    "borg",
    "bose",
    "bouman",
    "boyd",
    "brahmagupta",
    "brattain",
    "brown",
    "buck",
    "burnell",
    "cannon",
    "carson",
    "cartwright",
    "carver",
    "cerf",
    "chandrasekhar",
    "chaplygin",
    "chatelet",
    "chatterjee",
    "chebyshev",
    "cohen",
    "chaum",
    "clarke",
    "colden",
    "cori",
    "cray",
    "curran",
    "curie",
    "darwin",
    "davinci",
    "dewdney",
    "dhawan",
    "diffie",
    "dijkstra",
    "dirac",
    "driscoll",
    "dubinsky",
    "easley",
    "edison",
    "einstein",
    "elbakyan",
    "elgamal",
    "elion",
    "ellis",
    "engelbart",
    "euclid",
    "euler",
    "faraday",
    "feistel",
    "fermat",
    "fermi",
    "feynman",
    "franklin",
    "gagarin",
    "galileo",
    "galois",
    "ganguly",
    "gates",
    "gauss",
    "germain",
    "goldberg",
    "goldstine",
    "goldwasser",
    "golick",
    "goodall",
    "gould",
    "greider",
    "grothendieck",
    "haibt",
    "hamilton",
    "haslett",
    "hawking",
    "hellman",
    "heisenberg",
    "hermann",
    "herschel",
    "hertz",
    "heyrovsky",
    "hodgkin",
    "hofstadter",
    "hoover",
    "hopper",
    "hugle",
    "hypatia",
    "ishizaka",
    "jackson",
    "jang",
    "jemison",
    "jennings",
    "jepsen",
    "johnson",
    "joliot",
    "jones",
    "kalam",
    "kapitsa",
    "kare",
    "keldysh",
    "keller",
    "kepler",
    "khayyam",
    "khorana",
    "kilby",
    "kirch",
    "knuth",
    "kowalevski",
    "lalande",
    "lamarr",
    "lamport",
    "leakey",
    "leavitt",
    "lederberg",
    "lehmann",
    "lewin",
    "lichterman",
    "liskov",
    "lovelace",
    "lumiere",
    "mahavira",
    "margulis",
    "matsumoto",
    "maxwell",
    "mayer",
    "mccarthy",
    "mcclintock",
    "mclaren",
    "mclean",
    "mcnulty",
    "mendel",
    "mendeleev",
    "meitner",
    "meninsky",
    "merkle",
    "mestorf",
    "mirzakhani",
    "moore",
    "morse",
    "murdock",
    "moser",
    "napier",
    "nash",
    "neumann",
    "newton",
    "nightingale",
    "nobel",
    "noether",
    "northcutt",
    "noyce",
    "panini",
    "pare",
    "pascal",
    "pasteur",
    "payne",
    "perlman",
    "pike",
    "poincare",
    "poitras",
    "proskuriakova",
    "ptolemy",
    "raman",
    "ramanujan",
    "ride",
    "montalcini",
    "ritchie",
    "rhodes",
    "robinson",
    "roentgen",
    "rosalind",
    "rubin",
    "saha",
    "sammet",
    "sanderson",
    "satoshi",
    "shamir",
    "shannon",
    "shaw",
    "shirley",
    "shockley",
    "shtern",
    "sinoussi",
    "snyder",
    "solomon",
    "spence",
    "stonebraker",
    "sutherland",
    "swanson",
    "swartz",
    "swirles",
    "taussig",
    "tereshkova",
    "tesla",
    "tharp",
    "thompson",
    "torvalds",
    "tu",
    "turing",
    "varahamihira",
    "vaughan",
    "visvesvaraya",
    "volhard",
    "villani",
    "wescoff",
    "wilbur",
    "wiles",
    "williams",
    "williamson",
    "wilson",
    "wing",
    "wozniak",
    "wright",
    "wu",
    "yalow",
    "yonath",
    "zhukovsky",
]


# we could replace these with some favorite AI / Privacy researches and engineers?


def random_name() -> str:
    left_i = randbelow(len(left_name) - 1)
    right_i = randbelow(len(right_name) - 1)
    return f"{left_name[left_i].capitalize()} {right_name[right_i].capitalize()}"


def inherit_tags(
    attr_path_and_name: str,
    result: object,
    self_obj: Optional[object],
    args: Union[tuple, list],
    kwargs: dict,
) -> None:
    tags = []
    if self_obj is not None and hasattr(self_obj, "tags"):
        tags.extend(list(self_obj.tags))  # type: ignore

    for arg in args:
        if hasattr(arg, "tags"):
            tags.extend([tag for tag in arg.tags if tag not in tags])

    for arg in kwargs.values():
        if hasattr(arg, "tags"):
            tags.extend([tag for tag in arg.tags if tag not in tags])

    # only generate new tags if the result actually inherit some tags.
    if tags:
        tags.append(attr_path_and_name.split(".")[-1])
        result.tags = tags  # type: ignore


def get_root_data_path() -> Path:
    # get the PySyft / data directory to share datasets between notebooks
    # on Linux and MacOS the directory is: ~/.syft/data"
    # on Windows the directory is: C:/Users/$USER/.syft/data

    data_dir = Path.home() / ".syft" / "data"

    os.makedirs(data_dir, exist_ok=True)
    return data_dir



# stdlib
import sys
from typing import Any
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft relative
from ... import deserialize
from ... import serialize
from ...core.common import UID
from ...core.common.serde.serializable import bind_protobuf
from ...proto.lib.python.int_pb2 import Int as Int_PB
from .primitive_factory import PrimitiveFactory
from .primitive_interface import PyPrimitive
from .types import SyPrimitiveRet


@bind_protobuf
class Int(int, PyPrimitive):
    def __new__(
        cls, value: Any = None, base: Any = 10, id: Optional[UID] = None
    ) -> "Int":
        if value is None:
            value = 0

        if isinstance(value, str):
            return int.__new__(cls, value, base)

        return int.__new__(cls, value)

    def __init__(self, value: Any = None, base: Any = 10, id: Optional[UID] = None):
        if value is None:
            value = 0

        int.__init__(value)

        self._id: UID = id if id else UID()

    @property
    def id(self) -> UID:
        """We reveal PyPrimitive.id as a property to discourage users and
        developers of Syft from modifying .id attributes after an object
        has been initialized.

        :return: returns the unique id of the object
        :rtype: UID
        """
        return self._id

    def upcast(self) -> int:
        return int(self)

    def __add__(self, other: Any) -> SyPrimitiveRet:
        res = super().__add__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __int__(self) -> int:
        res = super().__int__()
        return res

    def __invert__(self) -> SyPrimitiveRet:
        res = super().__invert__()
        return PrimitiveFactory.generate_primitive(value=res)

    def __abs__(self) -> SyPrimitiveRet:
        res = super().__abs__()
        return PrimitiveFactory.generate_primitive(value=res)

    def __bool__(self) -> bool:
        return super().__bool__()

    def __divmod__(self, other: Any) -> SyPrimitiveRet:
        res = super().__divmod__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __rdivmod__(self, other: Any) -> SyPrimitiveRet:
        res = super().__rdivmod__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __radd__(self, other: Any) -> SyPrimitiveRet:
        res = super().__radd__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __sub__(self, other: Any) -> SyPrimitiveRet:
        res = super().__sub__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __rsub__(self, other: Any) -> SyPrimitiveRet:
        res = super().__rsub__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __rtruediv__(self, other: Any) -> SyPrimitiveRet:
        res = super().__rtruediv__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __mul__(self, other: Any) -> SyPrimitiveRet:
        res = super().__mul__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __rmul__(self, other: Any) -> SyPrimitiveRet:
        res = super().__rmul__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __ceil__(self) -> SyPrimitiveRet:
        res = super().__ceil__()
        return PrimitiveFactory.generate_primitive(value=res)

    def __eq__(self, other: Any) -> SyPrimitiveRet:
        res = super().__eq__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __float__(self) -> SyPrimitiveRet:
        res = super().__float__()
        return PrimitiveFactory.generate_primitive(value=res)

    def __floor__(self) -> SyPrimitiveRet:
        res = super().__floor__()
        return PrimitiveFactory.generate_primitive(value=res)

    def __floordiv__(self, other: Any) -> SyPrimitiveRet:
        res = super().__floordiv__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __rfloordiv__(self, other: Any) -> SyPrimitiveRet:
        res = super().__rfloordiv__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __truediv__(self, other: Any) -> SyPrimitiveRet:
        res = super().__truediv__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __mod__(self, other: Any) -> SyPrimitiveRet:
        res = super().__mod__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __rmod__(self, other: Any) -> SyPrimitiveRet:
        res = super().__rmod__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __pow__(self, other: Any, modulo: Optional[Any] = None) -> SyPrimitiveRet:
        if modulo:
            return PrimitiveFactory.generate_primitive(
                value=super().__pow__(other, modulo)
            )
        return PrimitiveFactory.generate_primitive(value=super().__pow__(other))

    def __rpow__(self, other: Any, modulo: Optional[Any] = None) -> SyPrimitiveRet:
        if modulo:
            return PrimitiveFactory.generate_primitive(
                value=super().__rpow__(other, modulo)
            )
        return PrimitiveFactory.generate_primitive(value=super().__rpow__(other))

    def __lshift__(self, other: Any) -> SyPrimitiveRet:
        res = super(Int, self).__lshift__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __rlshift__(self, other: Any) -> SyPrimitiveRet:
        res = super().__rlshift__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __round__(self, ndigits: Any = 0) -> SyPrimitiveRet:
        res = super().__round__(ndigits)
        return PrimitiveFactory.generate_primitive(value=res)

    def __rshift__(self, other: Any) -> SyPrimitiveRet:
        res = super(Int, self).__rshift__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __rrshift__(self, other: Any) -> SyPrimitiveRet:
        res = super().__rrshift__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __and__(self, other: Any) -> SyPrimitiveRet:
        res = super().__and__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __rand__(self, other: Any) -> SyPrimitiveRet:
        res = super().__rand__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __xor__(self, other: Any) -> SyPrimitiveRet:
        res = super(Int, self).__xor__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __rxor__(self, other: Any) -> SyPrimitiveRet:
        res = super().__rxor__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __or__(self, other: Any) -> SyPrimitiveRet:
        res = super().__or__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __ror__(self, other: Any) -> SyPrimitiveRet:
        res = super().__ror__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __ge__(self, other: Any) -> SyPrimitiveRet:
        res = super().__ge__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __lt__(self, other: Any) -> SyPrimitiveRet:
        res = super().__lt__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __le__(self, other: Any) -> SyPrimitiveRet:
        res = super().__le__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __gt__(self, other: Any) -> SyPrimitiveRet:
        res = super().__gt__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __iadd__(self, other: Any) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(
            value=super().__add__(other), id=self.id
        )

    def __isub__(self, other: Any) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(
            value=super().__sub__(other), id=self.id
        )

    def __imul__(self, other: Any) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(
            value=super().__mul__(other), id=self.id
        )

    def __ifloordiv__(self, other: Any) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(
            value=super().__floordiv__(other), id=self.id
        )

    def __itruediv__(self, other: Any) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(
            value=super().__truediv__(other), id=self.id
        )

    def __imod__(self, other: Any) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(
            value=super().__mod__(other), id=self.id
        )

    def __ipow__(self, other: Any, modulo: Optional[Any] = None) -> SyPrimitiveRet:
        if modulo:
            PrimitiveFactory.generate_primitive(
                value=super().__pow__(other, modulo), id=self.id
            )
        return PrimitiveFactory.generate_primitive(
            value=super().__pow__(other), id=self.id
        )

    def __ne__(self, other: Any) -> SyPrimitiveRet:
        res = super().__ne__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __hash__(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=super().__hash__())

    # add tests

    def __neg__(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=super().__neg__())

    def __pos__(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=super().__pos__())

    def _object2proto(self) -> Int_PB:
        int_pb = Int_PB()
        int_pb.data = self
        int_pb.id.CopyFrom(serialize(obj=self.id))
        return int_pb

    @staticmethod
    def _proto2object(proto: Int_PB) -> "Int":
        int_id: UID = deserialize(blob=proto.id)

        de_int = Int(value=proto.data)
        de_int._id = int_id  # can't use uid=int_id for some reason

        return de_int

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return Int_PB

    def as_integer_ratio(self) -> SyPrimitiveRet:
        if sys.version_info < (3, 8):
            raise NotImplementedError
        else:
            tpl = super().as_integer_ratio()
            return PrimitiveFactory.generate_primitive(value=tpl)

    def bit_length(self) -> SyPrimitiveRet:
        res = super().bit_length()
        return PrimitiveFactory.generate_primitive(value=res)

    def denominator(self) -> SyPrimitiveRet:
        res = super().denominator
        return PrimitiveFactory.generate_primitive(value=res)

    def to_bytes(
        self,
        length: int,
        byteorder: str,
        signed: bool = False,
    ) -> bytes:
        return int.to_bytes(self, length=length, byteorder=byteorder, signed=signed)

    @staticmethod
    def from_bytes(bytes: Any, byteorder: str, *, signed: Any = True) -> SyPrimitiveRet:
        res = int.from_bytes(bytes, byteorder, signed=signed)
        return PrimitiveFactory.generate_primitive(value=res)

    def imag(self) -> SyPrimitiveRet:
        res = super().imag
        return PrimitiveFactory.generate_primitive(value=res)

    def numerator(self) -> int:
        res = super().numerator
        return PrimitiveFactory.generate_primitive(value=res)

    def real(self) -> int:
        res = super().real
        return PrimitiveFactory.generate_primitive(value=res)

    def conjugate(self) -> SyPrimitiveRet:
        res = super().conjugate()
        return PrimitiveFactory.generate_primitive(value=res)
